//
//  DialPadButton.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import UIKit

@IBDesignable public class DialPadButton: UIButton {
    
    static let DefaultIndentWidth: CGFloat = 30
    static let DefaultNumberFontSize: CGFloat = 45
    static let DefaultAccentHeight: CGFloat = 4
    static let DefaultAccentBottomMargin: CGFloat = 11
    static let DefaultPrimaryColor: UIColor = UIColor.blackColor()
    
    @IBInspectable public var indentWidth: CGFloat {
        didSet {
            indentTextWidthConstraint?.constant = indentWidth
            let hasIndentText = indentText != nil && !indentText!.isEmpty
            let img = imageForState(UIControlState.Normal)
            let hasImage = img != nil
            var insets = self.titleEdgeInsets
            insets.left = hasIndentText ? indentWidth : hasImage ? indentWidth - img!.size.width : 0
            self.titleEdgeInsets = insets
        }
    }
    
    @IBInspectable public var indentText: String? {
        didSet {
            number?.text = indentText
            let img = imageForState(UIControlState.Normal)
            let hasImage = img != nil
            let hasIndentText = indentText != nil && !indentText!.isEmpty
            accent.hidden = !hasIndentText && !hasImage
            let inset = hasIndentText ? indentWidth : hasImage ? indentWidth - img!.size.width : 0
            var insets = self.titleEdgeInsets
            insets.left = inset
            self.titleEdgeInsets = insets
            accentBottomMarginConstraint.constant = hasIndentText ? DialPadButton.DefaultAccentBottomMargin : 0
        }
    }
    
    @IBInspectable public var primaryColor: UIColor? {
        didSet {
            number?.textColor = primaryColor
            accent?.backgroundColor = primaryColor
        }
    }
    
    override public var highlighted: Bool {
        didSet {
            configColor()
        }
    }
    
    override public var selected: Bool {
        didSet {
            configColor()
        }
    }
    
    override public func setImage(image: UIImage?, forState state: UIControlState) {
        super.setImage(image, forState: state)
        
        if state == .Normal {
            let hasImage = image != nil
            let hasIndentText = indentText != nil && !indentText!.isEmpty
            accent.hidden = !hasIndentText && !hasImage
            let inset = hasIndentText ? indentWidth : hasImage ? indentWidth - image!.size.width : 0
            var insets = self.titleEdgeInsets
            insets.left = inset
            self.titleEdgeInsets = insets
        }
    }
    
    private var number: UILabel!
    private var accent: UIView!
    private var indentTextWidthConstraint: NSLayoutConstraint!
    private var accentBottomMarginConstraint: NSLayoutConstraint!
    
    required public init?(coder aDecoder: NSCoder) {
        indentWidth = DialPadButton.DefaultIndentWidth
        super.init(coder: aDecoder)
        setup()
    }
    
    required override public init(frame: CGRect) {
        indentWidth = DialPadButton.DefaultIndentWidth
        super.init(frame: frame)
        setup()
    }
    
    private func configColor() {
        let color = highlighted ? self.titleColorForState(UIControlState.Highlighted) : selected ? self.titleColorForState(UIControlState.Selected) : primaryColor
        number?.textColor = color
        accent?.backgroundColor = color
        
        if highlighted || selected {
            layer.shadowColor = color?.CGColor
            layer.shadowOpacity = 1
        } else {
            layer.shadowOpacity = 0
        }
    }
    
    private func setup() {
        number = UILabel(frame: CGRectZero)
        number.translatesAutoresizingMaskIntoConstraints = false
        number.textAlignment = .Center
        number.textColor = DialPadButton.DefaultPrimaryColor
        number.font = UIFont(name: number.font.fontName, size: DialPadButton.DefaultNumberFontSize)
        addSubview(number)
        accent = UIView(frame: CGRectZero)
        accent.translatesAutoresizingMaskIntoConstraints = false
        accent.backgroundColor = DialPadButton.DefaultPrimaryColor
        addSubview(accent)
        
        let dict = [ "number" : number, "accent": accent, "view": self ]
        
        let hspacer = NSLayoutConstraint.constraintsWithVisualFormat("|[number]", options: [], metrics: nil, views: dict)
        let vpad = NSLayoutConstraint.constraintsWithVisualFormat("V:|[number]|", options: [], metrics: nil, views: dict)
        addConstraints(hspacer)
        addConstraints(vpad)
        indentTextWidthConstraint = NSLayoutConstraint(item: number, attribute: NSLayoutAttribute.Width, relatedBy: NSLayoutRelation.Equal, toItem: nil, attribute: NSLayoutAttribute.NotAnAttribute, multiplier: 1, constant: indentWidth)
        addConstraint(indentTextWidthConstraint)
        
        let hlayout = NSLayoutConstraint.constraintsWithVisualFormat("[number][accent]|", options: [], metrics: nil, views: dict)
        let vlayout = NSLayoutConstraint.constraintsWithVisualFormat("V:[accent(\(DialPadButton.DefaultAccentHeight))]", options: [], metrics: nil, views: dict)
        addConstraints(hlayout)
        addConstraints(vlayout)
        accentBottomMarginConstraint = NSLayoutConstraint(item: self, attribute: NSLayoutAttribute.Bottom, relatedBy: NSLayoutRelation.Equal,
            toItem: accent, attribute: NSLayoutAttribute.Bottom, multiplier: 1, constant: DialPadButton.DefaultAccentBottomMargin)
        addConstraint(accentBottomMarginConstraint)
        
        // Reassign indentText to make sure we go through the didSet hook
        let text = indentText
        indentText = text
        
        layer.shadowRadius = 8
        layer.shadowOpacity = 0
        layer.shadowOffset = CGSize(width: 0, height: 0)
        layer.masksToBounds = false
    }
    
}
