//
//  CallLogEntryViewController.swift
//  VMS
//
//  Copyright © 2015 Skymotion. All rights reserved.
//

import UIKit
import CoreData
import AddressBook
import AddressBookUI
import ContactsUI

class CallLogEntryViewController: UITableViewController, ABNewPersonViewControllerDelegate, ABUnknownPersonViewControllerDelegate, CNContactViewControllerDelegate {
    
    static let UnknownCallerRowNum = 4
    static let KnownCallerRowNum = 3
    
    static let CallTitleRow = 0
    static let CallDetailsRow = 1
    static let CallNumberRow = 2
    static let SaveContactRow = 3
    
    var call: Call!
    var unknown: Bool = true
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        if call?.match?.contact == nil {
            unknown = true
            navigationItem.rightBarButtonItem = nil
        } else {
            unknown = false
        }
    }
    
    @IBAction func editContact(sender: UIBarButtonItem) {
        guard let contact = call?.match?.contact else {
            return
        }
        if #available(iOS 9, *) {
            let (cstore, ctact) = AddressBook.sharedInstance.contactForContactCache(contact)
            guard let c = ctact else {
                return
            }
            
            editContact(cstore, contact: c)
        } else {
            let (abook, record) = AddressBook.sharedInstance.recordForContactCache(contact)
            guard let rec = record else {
                return
            }
            guard let ab = abook else {
                return
            }
            
            editContact(ab, record: rec)
        }
    }
    
    // MARK: - UITableViewDataSource
    override func numberOfSectionsInTableView(tableView: UITableView) -> Int {
        return 1
    }
    
    override func tableView(tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return unknown ? CallLogEntryViewController.UnknownCallerRowNum : CallLogEntryViewController.KnownCallerRowNum
    }
    
    override func tableView(tableView: UITableView, heightForRowAtIndexPath indexPath: NSIndexPath) -> CGFloat {
        switch indexPath.row {
        case CallLogEntryViewController.CallTitleRow:
            return 64.0
        case CallLogEntryViewController.CallDetailsRow:
            return 56.0
        case CallLogEntryViewController.CallNumberRow:
            if unknown {
                return tableView.rowHeight
            } else {
                return 64.0
            }
        default:
            return tableView.rowHeight
        }
    }
    
    override func tableView(tableView: UITableView, cellForRowAtIndexPath indexPath: NSIndexPath) -> UITableViewCell {
        
        let cell: UITableViewCell!
        
        switch indexPath.row {
        case CallLogEntryViewController.CallTitleRow:
            let tcell = tableView.dequeueReusableCellWithIdentifier("CallLogEntryTitleCellView", forIndexPath: indexPath) as! CallLogEntryTitleCellView
            tcell.nameOrPhone.text = call.displayName()
            
            cell = tcell
            break;
        case CallLogEntryViewController.CallDetailsRow:
            let dcell = tableView.dequeueReusableCellWithIdentifier("CallLogEntryDetailsCellView", forIndexPath: indexPath) as! CallLogEntryDetailsCellView
            
            dcell.date.text = call.displayDate()
            dcell.time.text = call.displayTime()
            dcell.callType.text = call.displayCallType()
            dcell.duration.text = call.displayDuration()
            
            cell = dcell
            break;
        case CallLogEntryViewController.CallNumberRow:
            if unknown {
                let acell = tableView.dequeueReusableCellWithIdentifier("CallLogEntryActionCellView", forIndexPath: indexPath) as! CallLogEntryActionCellView
                acell.actionLabel.text = NSLocalizedString("calllog.action.call", comment: "")
                
                cell = acell
            } else {
                let ncell = tableView.dequeueReusableCellWithIdentifier("CallLogEntryNumberCellView", forIndexPath: indexPath) as! CallLogEntryNumberCellView
                ncell.numberType.text = call.displayLabel()
                ncell.number.text = call.match?.number
                cell = ncell
            }
            break;
        case CallLogEntryViewController.SaveContactRow:
            let acell = tableView.dequeueReusableCellWithIdentifier("CallLogEntryActionCellView", forIndexPath: indexPath) as! CallLogEntryActionCellView
            acell.actionLabel.text = NSLocalizedString("calllog.action.save.contact", comment: "")
            
            cell = acell
            break;
        default:
            cell = nil
            break;
        }
        
        return cell
    }
    
    // MARK: - UITableViewDelegate
    override func tableView(tableView: UITableView, shouldHighlightRowAtIndexPath indexPath: NSIndexPath) -> Bool {
        switch indexPath.row {
        case CallLogEntryViewController.CallTitleRow:
            return false
        case CallLogEntryViewController.CallDetailsRow:
            return false
        default:
            return true
        }
    }
    
    override func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
        tableView.deselectRowAtIndexPath(indexPath, animated: true)
        
        switch indexPath.row {
        case CallLogEntryViewController.CallNumberRow:
            startCall(call.displayNumber()!, displayType: call?.match?.displayLabel(), displayName: call?.match?.contact?.displayFullName())
            break;
        case CallLogEntryViewController.SaveContactRow:
            if #available(iOS 9, *) {
                if let number = call.displayNumber() {
                    let contact = CNMutableContact()
                    contact.phoneNumbers = [CNLabeledValue(
                        label:CNLabelPhoneNumberMobile,
                        value:CNPhoneNumber(stringValue: number))]
                    
                    let vc = CNContactViewController(forUnknownContact: contact)
                    vc.contactStore = CNContactStore()
                    vc.delegate = self
                    vc.allowsActions = false
                    navigationController?.pushViewController(vc, animated: true)
                }
            } else {
                let record: ABRecord = ABPersonCreate().takeRetainedValue() as ABRecord
                let mvalue: ABMultiValue = ABMultiValueCreateMutable(ABPropertyType(kABMultiStringPropertyType)).takeRetainedValue() as ABMultiValue
                ABMultiValueAddValueAndLabel(mvalue, call.displayNumber(), kABPersonPhoneMobileLabel, nil)
                var error: Unmanaged<CFError>? = nil
                if ABRecordSetValue(record, kABPersonPhoneProperty, mvalue, &error) {
                    let vc = ABUnknownPersonViewController()
                    vc.allowsAddingToAddressBook = true
                    vc.allowsActions = false
                    vc.displayedPerson = record
                    navigationController?.pushViewController(vc, animated: true)
                }
            }

            break;
        default:
            break;
        }
    }
    
    // MARK: - ABNewPersonViewControllerDelegate
    func newPersonViewController(newPersonView: ABNewPersonViewController, didCompleteWithNewPerson person: ABRecord?) {
        newPersonView.dismissViewControllerAnimated(true, completion: {})
    }
    
    // MARK: - ABUnknownPersonViewControllerDelegate
    func unknownPersonViewController(personViewController: ABUnknownPersonViewController, shouldPerformDefaultActionForPerson person: ABRecord, property: ABPropertyID, identifier: ABMultiValueIdentifier) -> Bool {
        return false
    }
    
    func unknownPersonViewController(unknownCardViewController: ABUnknownPersonViewController, didResolveToPerson person: ABRecord?) {
        navigationController?.popViewControllerAnimated(true)
    }
    
    // MARK: - CNContactViewControllerDelegate
    @available(iOS 9.0, *)
    func contactViewController(viewController: CNContactViewController, shouldPerformDefaultActionForContactProperty property: CNContactProperty) -> Bool {
        return callContact(viewController.contact, property: property)
    }
    
    @available(iOS 9.0, *)
    func contactViewController(viewController: CNContactViewController, didCompleteWithContact contact: CNContact?) {
        if presentedViewController != nil {
            dismissViewControllerAnimated(true, completion: {})
        }
    }
}
