//
//  CallLogViewController.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import UIKit
import CoreData

class CallLogViewController: UIViewController, UITableViewDataSource, UITableViewDelegate, NSFetchedResultsControllerDelegate {
    
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var callFilter: UISegmentedControl!
    @IBOutlet var deleteButton: UIBarButtonItem!
    
    var callsFetchController: NSFetchedResultsController!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        navigationItem.rightBarButtonItem = editButtonItem()
        navigationItem.leftBarButtonItem = nil

        setupCallsFetchController()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    override func setEditing(editing: Bool, animated: Bool) {
        super.setEditing(editing, animated: animated)
        tableView.setEditing(editing, animated: animated)
        navigationItem.leftBarButtonItem = editing ? deleteButton : nil
    }
    
    override func prepareForSegue(segue: UIStoryboardSegue, sender: AnyObject?) {
        if segue.identifier == "CallLogEntryViewController" {
            let indexPath = tableView.indexPathForCell(sender as! UITableViewCell)
            let call = callsFetchController.objectAtIndexPath(indexPath!) as! Call
            let vc = segue.destinationViewController as! CallLogEntryViewController
            vc.call = call
        }
    }
    
    private func setupCallsFetchController() {
        setupCallsFetchController(false)
    }
    private func setupCallsFetchController(missedOnly: Bool) {
        callsFetchController = CallLog.sharedInstance.callsFetchController(missedOnly)
        callsFetchController.delegate = self
        var error: NSError? = nil
        do {
            try callsFetchController.performFetch()
        } catch let error1 as NSError {
            error = error1
            DDLogError("Error fetching call list: \(error), \(error?.userInfo)")
        }
    }
    
    @IBAction func callFilterChanged(sender: UISegmentedControl) {
        setupCallsFetchController(sender.selectedSegmentIndex > 0)
        tableView.reloadData()
    }
    
    @IBAction func deleteAll(sender: UIBarButtonItem) {
        CallLog.sharedInstance.deleteAll()
        tableView.editing = false
    }
    // MARK: - NSFetchedResultsControllerDelegate
    
    func controllerWillChangeContent(controller: NSFetchedResultsController) {
        DDLogVerbose("controllerWillChangeContent")
        tableView.beginUpdates()
    }
    
    func controller(controller: NSFetchedResultsController, didChangeSection sectionInfo: NSFetchedResultsSectionInfo, atIndex sectionIndex: Int, forChangeType type: NSFetchedResultsChangeType) {
        DDLogVerbose("didChangeSection \(sectionIndex)")
        switch(type) {
        case NSFetchedResultsChangeType.Insert:
            tableView.insertSections(NSIndexSet(index: sectionIndex), withRowAnimation: .Fade)
        case NSFetchedResultsChangeType.Delete:
            tableView.deleteSections(NSIndexSet(index: sectionIndex), withRowAnimation: .Fade)
        default:
            break
        }
    }
    
    func controller(controller: NSFetchedResultsController, didChangeObject anObject: AnyObject, atIndexPath indexPath: NSIndexPath?, forChangeType type: NSFetchedResultsChangeType, newIndexPath: NSIndexPath?) {
        DDLogVerbose("didChangeObject \(indexPath)")
        switch(type) {
        case NSFetchedResultsChangeType.Insert:
            tableView.insertRowsAtIndexPaths([ newIndexPath! ], withRowAnimation: .Automatic)
        case NSFetchedResultsChangeType.Delete:
            tableView.deleteRowsAtIndexPaths([ indexPath! ], withRowAnimation: .Automatic)
        case NSFetchedResultsChangeType.Update:
            tableView.reloadRowsAtIndexPaths([ indexPath! ], withRowAnimation: .None)
        case NSFetchedResultsChangeType.Move:
            tableView.moveRowAtIndexPath(indexPath!, toIndexPath: newIndexPath!)
        }
    }
    
    func controllerDidChangeContent(controller: NSFetchedResultsController) {
        DDLogVerbose("controllerDidChangeContent")
        tableView.endUpdates()
    }
    
    // MARK: - UITableViewDataSource
    
    func numberOfSectionsInTableView(tableView: UITableView) -> Int {
        if let sections = callsFetchController.sections {
            return sections.count
        } else {
            return 0
        }
    }
    
    func tableView(tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if let sect = callsFetchController.sections {
            if sect.count > section {
                let sinfo = sect[section] 
                return sinfo.numberOfObjects
            }
        }
        
        return 0
    }
    
    func tableView(tableView: UITableView, cellForRowAtIndexPath indexPath: NSIndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCellWithIdentifier("CallsListCell", forIndexPath: indexPath) as! CallLogCellView
        
        let call = callsFetchController.objectAtIndexPath(indexPath) as! Call
        
        cell.name.text = call.displayName()
        cell.numberType.text = call.displayLabel()
        cell.date.text = call.localizedDate()
        cell.outgoing.hidden = call.incoming.boolValue
        cell.name.textColor = call.missed ? UIColor.redColor() : UIColor.blackColor()
        
        return cell
    }
    
    // MARK: - UITableViewDelegate
    
    func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
        tableView.deselectRowAtIndexPath(indexPath, animated: true)
        
        let call = callsFetchController.objectAtIndexPath(indexPath) as? Call
        
        if let number = call?.displayNumber() {
            startCall(number, displayType: call?.match?.displayLabel(), displayName: call?.match?.contact?.displayFullName())
        }
    }
    
    func tableView(tableView: UITableView, canEditRowAtIndexPath indexPath: NSIndexPath) -> Bool {
        return true
    }
    
    func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
        if editingStyle == .Delete {
            let call = callsFetchController.objectAtIndexPath(indexPath) as! Call
            CallLog.sharedInstance.deleteCall(call)
        }
    }
}
