//
//  ContactListViewController.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import UIKit
import AddressBook
import AddressBookUI
import ContactsUI
import CoreData

class ContactListViewController: UIViewController, UITableViewDataSource, UITableViewDelegate, UISearchBarDelegate, UISearchResultsUpdating, NSFetchedResultsControllerDelegate, ABPersonViewControllerDelegate, ABNewPersonViewControllerDelegate, CNContactViewControllerDelegate {
    
    static let DefaultSectionDisplay = "#"
    
    @IBOutlet weak var tableView: UITableView!
    var searchController: UISearchController!
    var contactsFetchController: NSFetchedResultsController!
    var boldNameFont: UIFont?
    
    private static let alphabet = NSCharacterSet.uppercaseLetterCharacterSet()
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        setupContactsFetchController()
        setupSearchController()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    private func setupSearchController() {
        searchController = UISearchController(searchResultsController: nil)
        searchController.searchResultsUpdater = self
        searchController.dimsBackgroundDuringPresentation = false
        let sb = searchController.searchBar
        sb.delegate = self
        definesPresentationContext = true
        sb.sizeToFit()
        tableView.tableHeaderView = sb
    }
    
    private func setupContactsFetchController() {
        setupContactsFetchController(nil)
    }
    
    private func setupContactsFetchController(search: String?) {
        contactsFetchController = DataManager.sharedInstance.contactCacheFetchController(search)
        contactsFetchController.delegate = self
        var error: NSError? = nil
        do {
            try contactsFetchController.performFetch()
        } catch let error1 as NSError {
            error = error1
            DDLogError("Error fetching contact list: \(error), \(error?.userInfo)")
        }
    }
    
    private func makeBoldNameFont(nameFont: UIFont) -> UIFont {
        
        if boldNameFont == nil {
            boldNameFont = nameFont.VMS_boldFont()
        }
        
        return boldNameFont!
    }
    
    @IBAction func addContact(sender: AnyObject) {
        let vc: UIViewController
        if #available(iOS 9, *) {
            let cvc = CNContactViewController(forNewContact: nil)
            cvc.contactStore = CNContactStore()
            cvc.delegate = self
            vc = cvc
        } else {
            let avc = ABNewPersonViewController()
            avc.newPersonViewDelegate = self
            vc = avc
        }
        let nav = UINavigationController(rootViewController: vc)
        presentViewController(nav, animated: true, completion: {})
    }
    
    // MARK: - NSFetchedResultsControllerDelegate
    
    func controllerWillChangeContent(controller: NSFetchedResultsController) {
        DDLogVerbose("controllerWillChangeContent")
        tableView.beginUpdates()
    }
    
    func controller(controller: NSFetchedResultsController, didChangeSection sectionInfo: NSFetchedResultsSectionInfo, atIndex sectionIndex: Int, forChangeType type: NSFetchedResultsChangeType) {
        DDLogVerbose("didChangeSection \(sectionIndex)")
        switch(type) {
        case NSFetchedResultsChangeType.Insert:
            tableView.insertSections(NSIndexSet(index: sectionIndex), withRowAnimation: .Fade)
        case NSFetchedResultsChangeType.Delete:
            tableView.deleteSections(NSIndexSet(index: sectionIndex), withRowAnimation: .Fade)
        default:
            break
        }
    }
    
    func controller(controller: NSFetchedResultsController, didChangeObject anObject: AnyObject, atIndexPath indexPath: NSIndexPath?, forChangeType type: NSFetchedResultsChangeType, newIndexPath: NSIndexPath?) {
        DDLogVerbose("didChangeObject \(indexPath)")
        switch(type) {
        case NSFetchedResultsChangeType.Insert:
            tableView.insertRowsAtIndexPaths([ newIndexPath! ], withRowAnimation: .Automatic)
        case NSFetchedResultsChangeType.Delete:
            tableView.deleteRowsAtIndexPaths([ indexPath! ], withRowAnimation: .Automatic)
        case NSFetchedResultsChangeType.Update:
            tableView.reloadRowsAtIndexPaths([ indexPath! ], withRowAnimation: .None)
        case NSFetchedResultsChangeType.Move:
            tableView.moveRowAtIndexPath(indexPath!, toIndexPath: newIndexPath!)
        }
    }
    
    func controllerDidChangeContent(controller: NSFetchedResultsController) {
        DDLogVerbose("controllerDidChangeContent")
        tableView.endUpdates()
    }
    
    // MARK: - UITableViewDataSource
    
    func numberOfSectionsInTableView(tableView: UITableView) -> Int {
        if let sections = contactsFetchController.sections {
            return sections.count
        } else {
            return 0
        }
    }
    
    func tableView(tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if let sect = contactsFetchController.sections {
            if sect.count > section {
                let sinfo = sect[section] 
                return sinfo.numberOfObjects
            }
        }
        
        return 0
    }
    
    func tableView(tableView: UITableView, cellForRowAtIndexPath indexPath: NSIndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCellWithIdentifier("ContactListCell", forIndexPath: indexPath) 

        let contact = contactsFetchController.objectAtIndexPath(indexPath) as! ContactCache
        cell.textLabel?.attributedText = contact.displayFullName(makeBoldNameFont(cell.textLabel!.font))
        
        return cell
    }
    
    func tableView(tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        if let sect = contactsFetchController.sections {
            if sect.count > section {
                let sinfo = sect[section] 
                if sinfo.name == AddressBook.DefaultSection {
                    return ContactListViewController.DefaultSectionDisplay
                } else {
                    return sinfo.name
                }
            }
        }
        
        return nil
    }
    
    func sectionIndexTitlesForTableView(tableView: UITableView) -> [String]? {
        var titles = contactsFetchController.sectionIndexTitles 
        if let index = titles.indexOf(AddressBook.DefaultSection) {
            titles[index] = ContactListViewController.DefaultSectionDisplay
        }
        
        return titles
    }
    
    func tableView(tableView: UITableView, sectionForSectionIndexTitle title: String, atIndex index: Int) -> Int {
        let mappedTitle: String
        if title == ContactListViewController.DefaultSectionDisplay {
            mappedTitle = AddressBook.DefaultSection
        } else {
            mappedTitle = title
        }
        return contactsFetchController.sectionForSectionIndexTitle(mappedTitle, atIndex: index)
    }

    // MARK: - UITableViewDelegate
    
    func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
        tableView.deselectRowAtIndexPath(indexPath, animated: true)
        
        let contact = contactsFetchController.objectAtIndexPath(indexPath) as! ContactCache
        
        if #available(iOS 9, *) {
            let (cstore, ctact) = AddressBook.sharedInstance.contactForContactCache(contact)
            if let c = ctact {
                let vc = CNContactViewController(forContact: c)
                vc.contactStore = cstore
                vc.delegate = self
                vc.allowsEditing = true
                vc.allowsActions = false
                self.navigationController?.pushViewController(vc, animated: true)
            }
        } else {
            let (abook, person) = AddressBook.sharedInstance.recordForContactCache(contact)
            
            if let record = person {
                let personController = ABPersonViewController()
                personController.personViewDelegate = self
                personController.addressBook = abook
                personController.displayedPerson = record
                personController.allowsEditing = true
                personController.allowsActions = false
                self.navigationController?.pushViewController(personController, animated: true)
            }
        }
    }
    
    func tableView(tableView: UITableView, canEditRowAtIndexPath indexPath: NSIndexPath) -> Bool {
        return true
    }
    
    func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
        if editingStyle == .Delete {
            let contact = contactsFetchController.objectAtIndexPath(indexPath) as! ContactCache
            AddressBook.sharedInstance.deleteContact(contact)
        }
    }
    
    // MARK: - UISearchResultsUpdating
    func updateSearchResultsForSearchController(searchController: UISearchController) {
        let searchString = searchController.searchBar.text
        setupContactsFetchController(searchString)
        contactsFetchController.delegate = self
        tableView.reloadData()
    }
    
    // MARK: - ABPersonViewControllerDelegate
    func personViewController(personViewController: ABPersonViewController, shouldPerformDefaultActionForPerson person: ABRecord, property: ABPropertyID, identifier: ABMultiValueIdentifier) -> Bool {
        return callPerson(person, property: property, identifier: identifier)
    }
    
    // MARK - ABNewPersonViewControllerDelegate
    func newPersonViewController(newPersonView: ABNewPersonViewController, didCompleteWithNewPerson person: ABRecord?) {
        newPersonView.dismissViewControllerAnimated(true, completion: {})
    }
    
    // MARK: - CNContactViewControllerDelegate
    @available(iOS 9.0, *)
    func contactViewController(viewController: CNContactViewController, shouldPerformDefaultActionForContactProperty property: CNContactProperty) -> Bool {
        return callContact(viewController.contact, property: property)
    }
    
    @available(iOS 9.0, *)
    func contactViewController(viewController: CNContactViewController, didCompleteWithContact contact: CNContact?) {
        if presentedViewController != nil {
            dismissViewControllerAnimated(true, completion: {})
        }
    }
}

