//
//  UserAccountViewController.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import UIKit

private var kvoContext = 0

class UserAccountViewController: UITableViewController, UITextFieldDelegate {
    
    @IBOutlet weak var usernameInput: UITextField!
    @IBOutlet weak var passwordInput: UITextField!
    @IBOutlet weak var serverInput: UITextField!
    @IBOutlet weak var registrationStatus: UILabel!
    @IBOutlet weak var waitAnim: UIActivityIndicatorView!
    
    var gateway: Gateway?
    
    private var username: String?
    private var password: String?
    private var server: String?
    private var observingStatus = false;
    private var inSetupWizard: Bool {
        get {
            return navigationController?.presentingViewController != nil
        }
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        if !inSetupWizard {
            navigationItem.leftBarButtonItem = nil
            navigationItem.rightBarButtonItem = nil
        } else if let vcs = navigationController?.viewControllers {
            if vcs.count > 0 && vcs[0] != self {
                navigationItem.leftBarButtonItem = nil
            }
        }
        
        observingStatus = true
        Phone.sharedInstance.addObserver(self, forKeyPath: "status", options: [], context: &kvoContext)
        
        loadSettings()
    }
    
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated)
        
        if isMovingToParentViewController() || isBeingPresented() {
            loadSettings()
        }
        updateRegistrationStatus()
    }
    
    override func viewWillDisappear(animated: Bool) {
        super.viewWillDisappear(animated)
        
        if isMovingFromParentViewController() || isBeingDismissed() {
            saveSettings()
        }
    }
    
    deinit {
        if observingStatus {
            Phone.sharedInstance.removeObserver(self, forKeyPath: "status")
        }
    }
    
    override func observeValueForKeyPath(keyPath: String?, ofObject object: AnyObject?, change: [String : AnyObject]?, context: UnsafeMutablePointer<Void>) {
        if context == &kvoContext {
            updateRegistrationStatus()
        }
    }
    
    @IBAction func cancelAction(sender: UIBarButtonItem) {
        dismiss()
    }
    
    @IBAction func doneAction(sender: UIBarButtonItem) {
        saveSettings()
        dismiss()
    }
    
    private func dismiss() {
        if let pres = navigationController?.presentingViewController {
            pres.dismissViewControllerAnimated(true, completion: {})
        } else {
            navigationController?.popViewControllerAnimated(true)
        }
    }
    
    func updateRegistrationStatus() {
        let label: String
        let anim: Bool
        switch Phone.sharedInstance.status {
        case .Registered:
            label = "settings.status.registered"
            anim = false
            break;
        case .Registering:
            label = "settings.status.registering"
            anim = true
            break;
        case .Unregistered:
            label = "settings.status.unregistered"
            anim = false
            break;
        case .Unregistering:
            label = "settings.status.unregistering"
            anim = true
            break;
        }
        
        registrationStatus.text = NSLocalizedString(label, comment: "")
        if anim {
            waitAnim.startAnimating()
        } else {
            waitAnim.stopAnimating()
        }
    }
    
    func loadSettings() {
        let phone = Phone.sharedInstance
        (username, password) = phone.getCredentials()
        server = phone.getServer()
        
        usernameInput.text = username
        passwordInput.text = password
        serverInput.text = server
    }
    
    func saveSettings() {
        let phone = Phone.sharedInstance
        
        let usr = usernameInput.text
        let pwd = passwordInput.text
        let svr = gateway?.address
        
        phone.setCredentials(usr, password: pwd, server: svr)
        if inSetupWizard && gateway != nil {
            phone.gateway = gateway!
        } else {
            phone.savePhoneConfig()
        }
        
        if username != usr
        || password != pwd
        || server != svr {
            phone.restart()
        }
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(textField: UITextField) -> Bool {
        if textField == usernameInput {
            passwordInput.becomeFirstResponder()
        } else if textField == passwordInput {
            serverInput.becomeFirstResponder()
        } else if textField == serverInput {
            textField.resignFirstResponder()
        }
        return true
    }
    
    // MARK: - UITableViewDataSource
    
    override func numberOfSectionsInTableView(tableView: UITableView) -> Int {
        let num = super.numberOfSectionsInTableView(tableView)
        if let gat = gateway {
            if gat.resolved {
                return num - 1
            }
        }
        return num
    }
    
    // MARK: - UITableViewDelegate
    
    override func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {        
        tableView.deselectRowAtIndexPath(indexPath, animated: true)
        
        let cell = tableView.cellForRowAtIndexPath(indexPath) as! TextSettingCellView
        
        cell.input.becomeFirstResponder()
    }
}
