//
//  AddressBook.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import AddressBook
import Contacts
import ContactsUI

private var _instance: AddressBook? = nil

class AddressBook: NSObject {
    static let DefaultSection = "|"
    static let ExistingBatchSize = 5
    static let DidChangeNotification = "VMS_AddressBookDidChangeNotification"
    
    @available(iOS 9.0, *)
    static let ContactKeysToFetch = [
        CNContactGivenNameKey,
        CNContactFamilyNameKey,
        CNContactPhoneNumbersKey
    ]
    
    private static let alphabet = NSCharacterSet.uppercaseLetterCharacterSet()
    var addressBook: RHAddressBook?
    private var backgroundQueue: dispatch_queue_t
    private static var leadingZeroRegexp: NSRegularExpression? = nil
    
    class var sharedInstance: AddressBook {
        if _instance == nil {
            _instance = AddressBook()
        }
        return _instance!
    }
    
    override init() {
        backgroundQueue = dispatch_queue_create("AddressBookBackgroundQueue", DISPATCH_QUEUE_SERIAL)
        super.init()
        
        if #available(iOS 9, *) {
            NSNotificationCenter.defaultCenter().addObserver(self, selector: "addressBookChanged:", name: CNContactStoreDidChangeNotification, object: nil)
            
            let status = CNContactStore.authorizationStatusForEntityType(.Contacts)
            
            switch status {
            case .Authorized:
                DDLogVerbose("Contacts already authorized")
                break;
            case .NotDetermined:
                DDLogVerbose("Contacts authorization status not determined")
                let cstore = CNContactStore()
                cstore.requestAccessForEntityType(.Contacts, completionHandler: { (granted: Bool, error: NSError?) in
                    if granted {
                        DDLogVerbose("Contacts authorization granted")
                        self.reloadContacts()
                    } else if let err = error {
                        DDLogError("Contacts authorization error: \(err), \(err.userInfo)")
                    } else {
                        DDLogVerbose("Contacts authorization DENIED")
                    }
                })
                break;
            default:
                break;
            }
        } else {
            NSNotificationCenter.defaultCenter().addObserver(self, selector: "addressBookChanged:", name: RHAddressBookExternalChangeNotification, object: nil)
            
            let status = RHAddressBook.authorizationStatus().rawValue
            
            switch (status) {
            case RHAuthorizationStatusAuthorized.rawValue:
                DDLogVerbose("Address Book already authorized")
                addressBook = RHAddressBook()
                break;
            case RHAuthorizationStatusNotDetermined.rawValue:
                DDLogVerbose("Address Book authorization status not determined")
                let ab = RHAddressBook()
                ab.requestAuthorizationWithCompletion({ (granted: Bool, error: NSError!) in
                    
                    if granted {
                        DDLogVerbose("Address Book authorization granted")
                        self.addressBook = ab
                        self.reloadContacts()
                    } else if let err = error {
                        DDLogError("Address Book authorization error: \(err), \(err.userInfo)")
                    } else {
                        DDLogVerbose("Address Book authorization DENIED")
                    }
                })
                break;
            default:
                break;
            }
        }
    }
    
    deinit {
       NSNotificationCenter.defaultCenter().removeObserver(self)
    }
    
    func reloadContacts() {
        dispatch_async(backgroundQueue, {
            DDLogVerbose("Reload contacts")
            if #available(iOS 9, *) {
                self.processContacts9()
            } else {
                if let contacts = self.addressBook?.people() {
                    self.processContacts(contacts)
                }
            }
            NSNotificationCenter.defaultCenter().postNotificationName(AddressBook.DidChangeNotification, object: self)
        })
    }
    
    func addressBookChanged(notification: NSNotification) {
        DDLogVerbose("Address Book changed")
        reloadContacts()
    }
    
    class func canonicalNumberForMatching(number: String) -> String {
        let phutil = NBPhoneNumberUtil.sharedInstance()
        var error: NSError? = nil
        var phnum: NBPhoneNumber? = NBPhoneNumber()
        var national: NSString? = nil
        do {
            let code = try phutil.maybeExtractCountryCode(number, metadata: nil, nationalNumber: &national, keepRawInput: true, phoneNumber: &phnum)
            if let err = error {
                DDLogError("Error attempting to extract Country code from number \(number): \(code), \(err), \(err.userInfo)")
            }
        } catch let error1 as NSError {
            error = error1
        }
        
        if national == nil {
            national = phutil.normalizePhoneNumber(number)
            if AddressBook.leadingZeroRegexp == nil {
                var error: NSError? = nil
                do {
                    let regexp = try NSRegularExpression(pattern: "^0+", options: [])
                    AddressBook.leadingZeroRegexp = regexp
                } catch let error1 as NSError {
                    error = error1
                    if let err = error {
                        DDLogError("Error creating leading zero regexp: \(err), \(err.userInfo)")
                    }
                }
            }
            
            if let regexp = AddressBook.leadingZeroRegexp {
                if let nat = national {
                    national = regexp.stringByReplacingMatchesInString(nat as String, options: [], range: NSMakeRange(0, nat.length), withTemplate: "")
                }
            }
            
        }
        
        if let nat = national {
            return nat as String
        } else {
            return number
        }
    }
    
    func matchPhoneNumber(number: String) -> PhoneNumberCache? {
        let cano = AddressBook.canonicalNumberForMatching(number)
        let matches = DataManager.sharedInstance.phoneNumberCacheMatchingNumber(cano)
        let phutil = NBPhoneNumberUtil.sharedInstance()
        
        for numb in matches {
            var error: NSError? = nil
            let match = phutil.isNumberMatch(number, second: numb.number, error: &error)
            if let err = error {
                DDLogError("Error trying to match numbers \(number) and \(numb.number): \(err), \(err.userInfo)")
            } else {
                switch match.rawValue {
                case NBEMatchTypeEXACT_MATCH.rawValue:
                    return numb
                case NBEMatchTypeNSN_MATCH.rawValue:
                    return numb
                case NBEMatchTypeSHORT_NSN_MATCH.rawValue:
                    return numb
                default:
                    break;
                }
            }
        }
        
        return nil
    }
    
    func matchPhoneNumberAsync(number: String, callback: (PhoneNumberCache?) -> Void) {
        dispatch_async(backgroundQueue, {
            if let pho = self.matchPhoneNumber(number) {
                let oid = pho.objectID
                dispatch_async(dispatch_get_main_queue(), {
                    let cache = DataManager.sharedInstance.managedObjectContext?.objectWithID(oid) as! PhoneNumberCache?
                    callback(cache)
                })
            } else {
                dispatch_async(dispatch_get_main_queue(), {
                    callback(nil)
                })
            }
        })
    }
    
    private class func nameInitial(name: String?) -> String? {
        var initial: String? = nil
        if let nme = name {
            if(!nme.isEmpty) {
                initial = nme.substringToIndex(nme.startIndex.advancedBy(1))
                if let firstLetter = initial {
                    initial = firstLetter.uppercaseString
                }
            }
        }
        
        return initial
    }
    
    private class func nameInitialSection(initial: String?) -> String {
        
        var section = AddressBook.DefaultSection
        
        if let letter = initial {
            if letter.rangeOfCharacterFromSet(AddressBook.alphabet, options: [], range: nil) != nil {
                /*
                * http://stackoverflow.com/questions/5050270/nsstring-how-to-go-from-%C3%81lgebra-to-algebra
                */
                if let asciiData = letter.dataUsingEncoding(NSASCIIStringEncoding, allowLossyConversion: true) {
                    if let asciiString = NSString(data: asciiData, encoding: NSASCIIStringEncoding) {
                        section = asciiString as String
                    }
                }
            }
        }
        
        return section
    }
    
    // MARK: ABAddressBook specific methods (iOS 8)
    
    func recordForContactCache(contact: ContactCache) -> (ABAddressBook?, ABRecord?) {
        var person: ABRecord? = nil
        
        let rid = contact.recordID.intValue
        var error: Unmanaged<CFError>?
        let ab = ABAddressBookCreateWithOptions(nil, &error).takeUnretainedValue()
        if let err = error?.takeRetainedValue() {
            DDLogError("Couldn't create AddressBook: \(err)")
        } else {
            if let per = ABAddressBookGetPersonWithRecordID(ab, rid) {
                person = per.takeUnretainedValue()
            }
        }
        
        return (ab, person)
    }
    
    func deleteContact(contact: ContactCache) -> Bool {
        
        if #available(iOS 9, *) {
            return deleteContact9(contact)
        }
        
        var result = false
        if let person = self.addressBook?.personForABRecordID(contact.recordID.intValue) {
            var error: NSError? = nil
            if let addr = self.addressBook {
                let ret: Bool
                do {
                    /*
                    * Strange syntax to silence warning:
                    * https://forums.developer.apple.com/thread/16248
                    */
                    try addr.removePerson(person, error: ())
                    ret = true
                } catch let error1 as NSError {
                    error = error1
                    ret = false
                }
                if let err = error {
                    DDLogError("Error deleting contact: \(err) \(err.userInfo)")
                }
                result = ret
                if result {
                    error = nil
                    do {
                        try addr.saveWithError()
                    } catch let error1 as NSError {
                        error = error1
                        if let err = error {
                            DDLogError("Error saving address book: \(err) \(err.userInfo)")
                        }
                    }
                }
            }
        }
        
        return result
    }
    
    private func processContacts(contacts: [AnyObject]) {
        let dm = DataManager.sharedInstance
        
        var existing = [ Int32: ContactCache ]()
        let len = contacts.count
        
        DDLogVerbose("Processing \(len) contacts")
        
        let all = dm.allContactCache()
        for cc in all {
            existing[cc.recordID.intValue] = cc
        }
        
        DDLogVerbose("Found \(existing.count) existing contacts")
        
        for obj in contacts {
            let contact = obj as! RHPerson
            let id = contact.recordID
            var contactCache: ContactCache
            
            if let cc = existing[id] {
                contactCache = cc
                existing.removeValueForKey(id)
            } else {
                contactCache = dm.makeContactCacheObject()
                contactCache.recordID = NSNumber(int: id)
            }
            AddressBook.fillContactCache(contactCache, contact: contact)
            AddressBook.createPhoneNumberCache(contactCache, contact: contact)
        }
        
        DDLogVerbose("Deleting old removed contacts")
        let val = existing.values
        for orphan in val {
            dm.deleteManagedObject(orphan, save: false)
        }
        DDLogVerbose("Saving deleted data")
        dm.save()
        DDLogVerbose("AddressBook Reload Contacts: Done")
    }
    
    private class func createPhoneNumberCache(contactCache: ContactCache, contact: RHPerson) {
        if let numbers = contact.phoneNumbers {
            let len = numbers.count()
            let dm = DataManager.sharedInstance
            let existing = contactCache.numbers.mutableCopy() as! NSMutableSet
            for(var i: UInt = 0; i < len; i++) {
                let label = numbers.labelAtIndex(i)
                let number = numbers.valueAtIndex(i) as! String
                
                var found: PhoneNumberCache? = nil
                for exNum in existing {
                    let num = exNum as! PhoneNumberCache
                    if num.label == label && num.number == number {
                        found = num
                        break
                    }
                }
                
                if let foundNumber = found {
                    existing.removeObject(foundNumber)
                } else {
                    let numberCache = dm.makePhoneNumberCacheObject()
                    numberCache.label = label
                    numberCache.number = number
                    numberCache.contact = contactCache
                    numberCache.canonicalNumber = canonicalNumberForMatching(number)
                }
            }
            
            for orphan in existing {
                let num = orphan as! PhoneNumberCache
                dm.deleteManagedObject(num, save: false)
            }
        }
    }
    
    private class func fillContactCache(contactCache: ContactCache, contact: RHPerson) {
        contactCache.firstname = contact.firstName
        contactCache.lastname = contact.lastName
        
        if let last = contact.lastName {
            if !last.isEmpty {
                contactCache.indexedName = last
                contactCache.section = AddressBook.nameInitialSection(AddressBook.nameInitial(last))
            }
        }
        
        if contactCache.indexedName == nil {
            contactCache.indexedName = contact.firstName
            contactCache.indexFirstname = NSNumber(bool: true)
            contactCache.section = AddressBook.nameInitialSection(AddressBook.nameInitial(contact.firstName))
        }
    }
    
    // MARK: Contacts framework methods (iOS 9+)
    
    @available(iOS 9, *)
    func contactForContactCache(contact: ContactCache) -> (CNContactStore, CNContact?) {
        
        let store = CNContactStore()
        var person: CNContact? = nil
        do {
            person = try store.unifiedContactWithIdentifier(contact.identifier, keysToFetch: [CNContactViewController.descriptorForRequiredKeys()])
        } catch let err as NSError {
            DDLogError("Error fetching contact: \(err) \(err.userInfo)")
        }
        
        return (store, person)
    }
    
    @available(iOS 9, *)
    func deleteContact9(contact: ContactCache) -> Bool {
        var result = false
        let store = CNContactStore()
        do {
            let person = try store.unifiedContactWithIdentifier(contact.identifier, keysToFetch: AddressBook.ContactKeysToFetch)
            let saveRequest = CNSaveRequest()
            saveRequest.deleteContact(person.mutableCopy() as! CNMutableContact)
            try store.executeSaveRequest(saveRequest)
            result = true
        } catch let err as NSError {
            DDLogError("Error deleting contact: \(err) \(err.userInfo)")
        }
        
        return result
    }
    
    @available(iOS 9, *)
    private func processContacts9() {
        let dm = DataManager.sharedInstance
        
        var existing = [ String: ContactCache ]()
        
        let all = dm.allContactCache()
        for cc in all {
            existing[cc.identifier] = cc
        }
        
        DDLogVerbose("Found \(existing.count) existing contacts")
        
        let request = CNContactFetchRequest(keysToFetch: AddressBook.ContactKeysToFetch)
        let store = CNContactStore()
        var len = 0
        do {
            try store.enumerateContactsWithFetchRequest(request, usingBlock: { (contact: CNContact, stop: UnsafeMutablePointer<ObjCBool>) in
                let id = contact.identifier
                var contactCache: ContactCache
                
                if let cc = existing[id] {
                    contactCache = cc
                    existing.removeValueForKey(id)
                } else {
                    contactCache = dm.makeContactCacheObject()
                    contactCache.identifier = id
                }
                AddressBook.fillContactCache(contactCache, contact: contact)
                AddressBook.createPhoneNumberCache(contactCache, contact: contact)
                len++
            })
        } catch let error as NSError {
            DDLogVerbose("Error enumerating contacts: \(error), \(error.userInfo)")
        }
        
        DDLogVerbose("Processed \(len) contacts")
        
        DDLogVerbose("Deleting old removed contacts")
        let val = existing.values
        for orphan in val {
            dm.deleteManagedObject(orphan, save: false)
        }
        DDLogVerbose("Saving deleted data")
        dm.save()
        DDLogVerbose("AddressBook Reload Contacts: Done")
    }
    
    @available(iOS 9, *)
    private class func createPhoneNumberCache(contactCache: ContactCache, contact: CNContact) {
        let numbers = contact.phoneNumbers
        let dm = DataManager.sharedInstance
        let existing = contactCache.numbers.mutableCopy() as! NSMutableSet
        for val in numbers {
            let label = val.label
            let numberVal = val.value as! CNPhoneNumber
            let number = numberVal.stringValue
            
            var found: PhoneNumberCache? = nil
            for exNum in existing {
                let num = exNum as! PhoneNumberCache
                if num.label == label && num.number == number {
                    found = num
                    break
                }
            }
            
            if let foundNumber = found {
                existing.removeObject(foundNumber)
            } else {
                let numberCache = dm.makePhoneNumberCacheObject()
                numberCache.label = label
                numberCache.number = number
                numberCache.contact = contactCache
                numberCache.canonicalNumber = canonicalNumberForMatching(number)
            }
        }
        
        for orphan in existing {
            let num = orphan as! PhoneNumberCache
            dm.deleteManagedObject(num, save: false)
        }
    }
    
    @available(iOS 9, *)
    private class func fillContactCache(contactCache: ContactCache, contact: CNContact) {
        let first = contact.givenName
        let last = contact.familyName
        contactCache.firstname = first
        contactCache.lastname = last
        
        if !last.isEmpty {
            contactCache.indexedName = last
            contactCache.section = AddressBook.nameInitialSection(AddressBook.nameInitial(last))
        }
        
        if contactCache.indexedName == nil {
            contactCache.indexedName = first
            contactCache.indexFirstname = NSNumber(bool: true)
            contactCache.section = AddressBook.nameInitialSection(AddressBook.nameInitial(first))
        }
    }
}
