//
//  CallLog.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import Foundation
import CoreData

class CallLog: NSObject {
    private static var _instance: CallLog? = nil
    private var pendingCalls = [Int: Call]()
    
    class var sharedInstance: CallLog {
        if _instance == nil {
            _instance = CallLog()
        }
        return _instance!
    }
    
    override init() {
        super.init()
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "addressBookDidChange:", name: AddressBook.DidChangeNotification, object: nil)
    }
    
    deinit {
        NSNotificationCenter.defaultCenter().removeObserver(self)
    }
    
    func startCall(callId: Int, number: String, incoming: Bool) {
        let call = DataManager.sharedInstance.makeUnmanagedCallObject()
        call.incoming = NSNumber(bool: incoming)
        call.date = NSDate()
        if number.hasPrefix("<") {
            call.number = number
        } else {
            call.number = "<\(number)>"
        }
        pendingCalls[callId] = call
    }
    
    func connectCall(callId: Int) {
        if let call = pendingCalls[callId] {
            call.connectDate = NSDate()
        }
    }
    
    func disconnectCall(callId: Int) {
        if let call = pendingCalls[callId] {
            if let connect = call.connectDate {
                let now = NSDate()
                call.duration = NSNumber(double: now.timeIntervalSinceDate(connect))
            }
            let dm = DataManager.sharedInstance
            dm.insertUnmanagedObject(call)
            pendingCalls.removeValueForKey(callId)
            dm.save()
            let oid = call.objectID
            let context = dm.backgroundObjectContext
            let phone = Phone.sharedInstance.phone
            context?.performBlock({
                if let bgCall = context?.objectWithID(oid) as? Call {
                    if let pnumber = AddressBook.sharedInstance.matchPhoneNumber(phone.sipUriUsername(bgCall.number)) {
                        bgCall.match = pnumber
                        dm.save()
                    }
                }
            })
        }
    }
    
    func addressBookDidChange(notification: NSNotification) {
        rematchNumbers()
    }
    
    func rematchNumbers() {
        let phone = Phone.sharedInstance.phone
        let dm = DataManager.sharedInstance
        let context = dm.backgroundObjectContext
        context?.performBlock({
            let orphans = self.orphanCalls()
            for call in orphans {
                if let pnumber = AddressBook.sharedInstance.matchPhoneNumber(phone.sipUriUsername(call.number)) {
                    call.match = pnumber
                }
            }
            if orphans.count > 0 {
                dm.save()
            }
        })
    }
    
    func orphanCalls() -> [Call] {
        return DataManager.sharedInstance.orphanCalls()
    }
    
    func callsFetchController(missedOnly: Bool) -> NSFetchedResultsController {
        return DataManager.sharedInstance.callsFetchController(missedOnly)
    }
    
    func allCallsFetchController() -> NSFetchedResultsController {
        return callsFetchController(false)
    }
    
    func missedCallsFetchController() -> NSFetchedResultsController {
        return callsFetchController(true)
    }
    
    func lastCalledNumber() -> String? {
        let call = DataManager.sharedInstance.lastOutgoingCall()
        
        return call?.displayNumber()
    }
    
    func deleteAll() {
        DataManager.sharedInstance.clearCalls()
    }
    
    func deleteCall(call: Call) {
        DataManager.sharedInstance.deleteManagedObject(call, save: true)
    }
}
