//
//  DataManager.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import CoreData

private let _instance = DataManager()

class DataManager: NSObject {
    
    static let ContactCacheEntity = "ContactCache"
    static let PhoneNumberCacheEntity = "PhoneNumberCache"
    static let CallEntity = "Call"
    
    class var sharedInstance: DataManager {
        return _instance
    }
    
    var saveBackgroundTaskId = UIBackgroundTaskInvalid
    var saveForegroundTaskId = UIBackgroundTaskInvalid
    
    func makeContactCacheObject() -> ContactCache {
        let contact = makeManagedObjectForEntity(DataManager.ContactCacheEntity)
        return contact as! ContactCache
    }

    func contactCacheFetchRequest() -> NSFetchRequest {
        return contactCacheFetchRequest(nil)
    }
    
    func existingContactCacheFetchRequest(context: NSManagedObjectContext, ids: [NSNumber]) -> NSFetchRequest {
        let entityDescription = NSEntityDescription.entityForName(DataManager.ContactCacheEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.predicate = NSPredicate(format: "recordID IN %@", ids)
        
        return request
    }
    
    func existingContactCache(context: NSManagedObjectContext, ids: [NSNumber]) -> [ContactCache] {
        let request = existingContactCacheFetchRequest(context, ids: ids)
        var results: [ContactCache]! = []
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let res = try context.executeFetchRequest(request)
                results = res as! [ContactCache]
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Failed to fetch existing Contact Cache: \(err.description)")
                }
            } catch {
                fatalError()
            }
        })
        
        return results
    }
    
    func existingContactCache(ids: [NSNumber]) -> [ContactCache] {
        return existingContactCache(objectContext, ids: ids)
    }

    func allContactCache() -> [ContactCache] {
        let context = objectContext
        let entityDescription = NSEntityDescription.entityForName(DataManager.ContactCacheEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        var results: [ContactCache]! = []
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let res = try context.executeFetchRequest(request)
                results = res as! [ContactCache]
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Failed to fetch all Contact Cache: \(err), \(err.userInfo)")
                }
            } catch {
                fatalError()
            }
        })
        
        return results
    }
    
    func deleteContactCacheWithTimestampBefore(timestamp: NSNumber) -> Int {
        return deleteEntityObjectsWithTimestampBefore(timestamp, entity: DataManager.ContactCacheEntity)
    }
    
    func contactCacheFetchRequest(search: String?) -> NSFetchRequest {
        let entityDescription = NSEntityDescription.entityForName(DataManager.ContactCacheEntity, inManagedObjectContext: managedObjectContext!)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.sortDescriptors = [ NSSortDescriptor(key: "section", ascending: true), NSSortDescriptor(key: "indexedName", ascending: true) ]
        request.fetchBatchSize = 20
        
        if let term = search {
            if !term.isEmpty {
                let fields = [ "firstname", "lastname" ]
                let variants = [ " %@", "-%@", "'%@" ]
                var predicates = [NSPredicate]()
                var beginPredicates = [NSPredicate]()
                var containsPredicates = [NSPredicate]()
                
                for field in fields {
                    let beginPredicate = NSPredicate(format: "\(field) BEGINSWITH[cd] %@", term)
                    beginPredicates.append(beginPredicate)
                    
                    for variant in variants {
                        let needle = String(format: variant, term)
                        let predicate = NSPredicate(format: "\(field) CONTAINS[cd] %@", needle)
                        containsPredicates.append(predicate)
                    }
                }
                predicates = beginPredicates + containsPredicates
                request.predicate = NSCompoundPredicate(type: .OrPredicateType, subpredicates: predicates)
            }
        }
        
        return request
    }

    func contactCacheFetchController() -> NSFetchedResultsController {
        return contactCacheFetchController(nil)
    }
    
    func contactCacheFetchController(search: String?) -> NSFetchedResultsController {
        let ctl = NSFetchedResultsController(fetchRequest: contactCacheFetchRequest(search), managedObjectContext: managedObjectContext!, sectionNameKeyPath: "section", cacheName: nil)
        
        return ctl
    }
    
    func clearContactCache() {
        clearContactCache(objectContext)
    }
    
    func clearContactCacheAndSave(save: Bool) {
        clearContactCache(objectContext, save: save)
    }

    func clearContactCache(context: NSManagedObjectContext) {
        clearContactCache(context, save: true)
    }
    
    func clearContactCache(context: NSManagedObjectContext, save: Bool) {
        clearEntityObjects(context, entity: DataManager.ContactCacheEntity, save: save)
    }
    
    func makePhoneNumberCacheObject() -> PhoneNumberCache {
        let contact = makeManagedObjectForEntity(DataManager.PhoneNumberCacheEntity)
        return contact as! PhoneNumberCache
    }
    
    func deletePhoneNumberCacheWithTimestampBefore(timestamp: NSNumber) -> Int {
        return deleteEntityObjectsWithTimestampBefore(timestamp, entity: DataManager.PhoneNumberCacheEntity)
    }
    
    func phoneNumberCacheMatchingNumber(number: String) -> [PhoneNumberCache] {
        let context = objectContext
        let entityDescription = NSEntityDescription.entityForName(DataManager.PhoneNumberCacheEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.predicate = NSPredicate(format: "canonicalNumber = %@", number)

        var results: [PhoneNumberCache]! = []
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let res = try context.executeFetchRequest(request)
                results = res as! [PhoneNumberCache]
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Failed to fetch Phone Numbers matching \(number): \(err), \(err.userInfo)")
                }
            } catch {
                fatalError()
            }
        })
        
        return results
    }
    
    func makeUnmanagedCallObject() -> Call {
        return makeUnmanagedObjectForEntity(DataManager.CallEntity) as! Call
    }
    
    func callsFetchRequest(missedOnly: Bool) -> NSFetchRequest {
        let context = objectContext
        let entityDescription = NSEntityDescription.entityForName(DataManager.CallEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription

        if missedOnly {
            request.predicate = NSPredicate(format: "incoming = %@ AND duration = 0", NSNumber(bool: true))
        }
        
        let sort = NSSortDescriptor(key: "date", ascending: false)
        request.sortDescriptors = [sort]
        request.fetchBatchSize = 20
        
        return request
    }
    
    func orphanCallsFetchRequest() -> NSFetchRequest {
        let context = objectContext
        let entityDescription = NSEntityDescription.entityForName(DataManager.CallEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.predicate = NSPredicate(format: "match == nil")
        
        return request
    }
    
    func orphanCalls() -> [Call] {
        let req = orphanCallsFetchRequest()
        var results = [Call]()
        let context = objectContext
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let res = try context.executeFetchRequest(req)
                results = res as! [Call]
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Error fetching orphan calls: \(err), \(err.userInfo)")
                }
            } catch {
                fatalError()
            }
        })
        
        return results
    }
    
    func lastOutgoingCall() -> Call? {
        let context = objectContext
        let entityDescription = NSEntityDescription.entityForName(DataManager.CallEntity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.predicate = NSPredicate(format: "incoming = %@", NSNumber(bool: false))
        request.fetchLimit = 1
        request.sortDescriptors = [ NSSortDescriptor(key: "date", ascending: false) ]
        
        var results = [Call]()
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let res = try context.executeFetchRequest(request)
                results = res as! [Call]
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Error fetching orphan calls: \(err), \(err.userInfo)")
                }
            } catch {
                fatalError()
            }
        })
        
        return results.count > 0 ? results[0] : nil
    }
    
    func callsFetchController(missedOnly: Bool) -> NSFetchedResultsController {
        let ctl = NSFetchedResultsController(fetchRequest: callsFetchRequest(missedOnly), managedObjectContext: managedObjectContext!, sectionNameKeyPath: nil, cacheName: nil)
        
        return ctl
    }
    
    func allCallsFetchController() -> NSFetchedResultsController {
        return callsFetchController(false)
    }
    
    func missedCallsFetchController() -> NSFetchedResultsController {
        return callsFetchController(true)
    }
    
    func clearCalls() {
        clearCalls(objectContext)
    }
    
    func clearCallsAndSave(save: Bool) {
        clearCalls(objectContext, save: save)
    }
    
    func clearCalls(context: NSManagedObjectContext) {
        clearCalls(context, save: true)
    }
    
    func clearCalls(context: NSManagedObjectContext, save: Bool) {
        clearEntityObjects(context, entity: DataManager.CallEntity, save: save)
    }
    
    func clearEntityObjects(context: NSManagedObjectContext, entity: String) {
        clearEntityObjects(context, entity: entity, save: true)
    }
    
    func clearEntityObjects(context: NSManagedObjectContext, entity: String, save: Bool) {
        context.performBlockAndWait({ () -> Void in
            let entityDescription = NSEntityDescription.entityForName(entity, inManagedObjectContext: context)
            let request = NSFetchRequest()
            request.entity = entityDescription
            request.includesPropertyValues = false
            
            var error: NSError? = nil
            do {
                let results = try context.executeFetchRequest(request)
                for contact in results {
                    let cont = contact as! NSManagedObject
                    DataManager.deleteManagedObject(context, object: cont, save: false)
                }
                if(save) {
                    self.save()
                }
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Failed to fetch all objects in entity \(entity): \(err.description)")
                }
            } catch {
                fatalError()
            }
        })
    }
    
    func deleteEntityObjectsWithTimestampBefore(timestamp: NSNumber, entity: String) -> Int {
        let context = objectContext
        
        let entityDescription = NSEntityDescription.entityForName(entity, inManagedObjectContext: context)
        let request = NSFetchRequest()
        request.entity = entityDescription
        request.predicate = NSPredicate(format: "timestamp < %@", timestamp)
        request.includesPropertyValues = false
        
        var num = 0
        context.performBlockAndWait({
            var error: NSError? = nil
            do {
                let results = try context.executeFetchRequest(request)
                for obj in results {
                    let cont = obj as! NSManagedObject
                    DataManager.deleteManagedObject(context, object: cont, save: false)
                }
                num = results.count
            } catch let error1 as NSError {
                error = error1
                if let err = error {
                    DDLogError("Failed to delete object of \(entity): \(err), \(err.userInfo)")
                }
            } catch {
                fatalError()
            }
        })
        
        return num
    }
    
    class func makeManagedObjectForEntity(context: NSManagedObjectContext, entity: String) -> NSManagedObject {
        return NSEntityDescription.insertNewObjectForEntityForName(entity, inManagedObjectContext: context) 
    }
    
    func makeManagedObjectForEntity(entity: String) -> NSManagedObject {
        return DataManager.makeManagedObjectForEntity(objectContext, entity: entity)
    }
    
    class func makeUnmanagedObjectForEntity(context: NSManagedObjectContext, entity: String) -> NSManagedObject {
        let desc = NSEntityDescription.entityForName(entity, inManagedObjectContext: context)
        return NSManagedObject(entity: desc!, insertIntoManagedObjectContext: nil)
    }
    
    func makeUnmanagedObjectForEntity(entity: String) -> NSManagedObject {
        return DataManager.makeUnmanagedObjectForEntity(objectContext, entity: entity)
    }
    
    class func insertUnmanagedObject(context: NSManagedObjectContext, object: NSManagedObject) {
        context.performBlockAndWait({
            context.insertObject(object)
        })
    }
    
    func insertUnmanagedObject(object: NSManagedObject) {
        DataManager.insertUnmanagedObject(objectContext, object: object)
    }
    
    class func deleteManagedObject(context: NSManagedObjectContext, object: NSManagedObject, save shouldSave: Bool ) {
        context.performBlockAndWait({ () -> Void in
            context.deleteObject(object)
        })
        if(shouldSave) {
            save(context: context)
        }
    }
    
    func deleteManagedObject(object: NSManagedObject, save shouldSave: Bool ) {
        return DataManager.deleteManagedObject(objectContext, object: object, save: shouldSave)
    }
    
    class func inverseRelationship(context: NSManagedObjectContext, entity: String, relationship: String) -> String {
        let desc : NSEntityDescription? = NSEntityDescription.entityForName(entity, inManagedObjectContext: context)
        let rel = desc!.relationshipsByName[relationship]
        if let name = rel?.inverseRelationship?.name {
            return name
        } else {
            return ""
        }
    }
    
    func inverseRelationship(entity: String, relationship: String) -> String {
        return DataManager.inverseRelationship(objectContext, entity: entity, relationship: relationship)
    }
    
    // MARK: - Core Data stack
    
    lazy var applicationDocumentsDirectory: NSURL = {
        // The directory the application uses to store the Core Data store file. This code uses a directory named "com.yellowbus.Courseboard" in the application's documents Application Support directory.
        let urls = NSFileManager.defaultManager().URLsForDirectory(.DocumentDirectory, inDomains: .UserDomainMask)
        return urls[urls.count-1] 
        }()
    
    lazy var managedObjectModel: NSManagedObjectModel = {
        // The managed object model for the application. This property is not optional. It is a fatal error for the application not to be able to find and load its model.
        let modelURL = NSBundle.mainBundle().URLForResource("vms", withExtension: "momd")!
        return NSManagedObjectModel(contentsOfURL: modelURL)!
        }()
    
    lazy var persistentStoreCoordinator: NSPersistentStoreCoordinator? = {
        // The persistent store coordinator for the application. This implementation creates and return a coordinator, having added the store for the application to it. This property is optional since there are legitimate error conditions that could cause the creation of the store to fail.
        // Create the coordinator and store
        var coordinator: NSPersistentStoreCoordinator? = NSPersistentStoreCoordinator(managedObjectModel: self.managedObjectModel)
        let url = self.applicationDocumentsDirectory.URLByAppendingPathComponent("vms.sqlite")
        var error: NSError? = nil
        var failureReason = "There was an error creating or loading the application's saved data."
        do {
            try coordinator!.addPersistentStoreWithType(NSSQLiteStoreType, configuration: nil, URL: url, options: nil)
        } catch var error1 as NSError {
            error = error1
            coordinator = nil
            // Report any error we got.
            let dict: [NSObject: AnyObject] = [
                NSLocalizedDescriptionKey: "Failed to initialize the application's saved data",
                NSLocalizedFailureReasonErrorKey: failureReason,
                NSUnderlyingErrorKey: error!
            ]
            error = NSError(domain: "YOUR_ERROR_DOMAIN", code: 9999, userInfo: dict)
            // Replace this with code to handle the error appropriately.
            // abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
            NSLog("Unresolved error \(error), \(error!.userInfo)")
            abort()
        } catch {
            fatalError()
        }
        
        return coordinator
        }()
    
    lazy var managedObjectContext: NSManagedObjectContext? = {
        // Returns the managed object context for the application (which is already bound to the persistent store coordinator for the application.) This property is optional since there are legitimate error conditions that could cause the creation of the context to fail.
            let coordinator = self.persistentStoreCoordinator
            if coordinator == nil {
                return nil
            }
            var managedObjectContext = NSManagedObjectContext(concurrencyType: .MainQueueConcurrencyType)
            managedObjectContext.persistentStoreCoordinator = coordinator
            return managedObjectContext
        }()
    
    lazy var backgroundObjectContext: NSManagedObjectContext? = {
        var context = NSManagedObjectContext(concurrencyType: .PrivateQueueConcurrencyType)
        context.parentContext = self.managedObjectContext!
        return context
    }()
    
    var objectContext: NSManagedObjectContext! {
        get {
            return NSThread.isMainThread() ? managedObjectContext : backgroundObjectContext
        }
    }
    
    // MARK: - Core Data Saving support
    
    class func save (context moc: NSManagedObjectContext) -> Bool {
        var error: NSError? = nil
        let res: Bool
        do {
            try moc.save()
            res = true
        } catch let error1 as NSError {
            error = error1
            res = false
        }
        if !res {
            DDLogError("Error saving context: \(error), \(error!.userInfo)")
        }
        
        return res
    }
    
    func save () {
        if NSThread.isMainThread() {
            saveForeground()
        } else {
            let app = UIApplication.sharedApplication()
            saveBackgroundTaskId = app.beginBackgroundTaskWithName("DataManager.save", expirationHandler: {
                UIApplication.sharedApplication().endBackgroundTask(self.saveBackgroundTaskId)
            })
            backgroundObjectContext?.performBlockAndWait( { () -> Void in
                if DataManager.save(context: self.backgroundObjectContext!) {
                    self.saveForeground()
                }
            })
            app.endBackgroundTask(saveBackgroundTaskId)
        }
    }
    
    func saveBackground() {
        let app = UIApplication.sharedApplication()
        saveBackgroundTaskId = app.beginBackgroundTaskWithName("DataManager.saveBackground", expirationHandler: {
            UIApplication.sharedApplication().endBackgroundTask(self.saveBackgroundTaskId)
        })
        backgroundObjectContext?.performBlockAndWait( { () -> Void in
            DataManager.save(context: self.backgroundObjectContext!)
        })
        app.endBackgroundTask(saveBackgroundTaskId)
        saveBackgroundTaskId = UIBackgroundTaskInvalid
    }
    
    func saveForeground() {
        let app = UIApplication.sharedApplication()
        saveForegroundTaskId = app.beginBackgroundTaskWithName("DataManager.saveForeground", expirationHandler: {
            UIApplication.sharedApplication().endBackgroundTask(self.saveForegroundTaskId)
        })
        managedObjectContext?.performBlockAndWait( { () -> Void in
            DataManager.save(context: self.managedObjectContext!)
        })
        app.endBackgroundTask(saveForegroundTaskId)
        saveForegroundTaskId = UIBackgroundTaskInvalid
    }
}
