//
//  Gateway.swift
//  VMS
//
//  Copyright © 2015 Skymotion. All rights reserved.
//

import Foundation

class Gateway: NSObject, NSNetServiceDelegate {
    
    static let ResolveTimeout: NSTimeInterval = 10.0
    
    dynamic var address: String? = nil
    dynamic var hostname: String? = nil
    dynamic var port: Int = 0
    
    var name: String? = nil
    var type: String? = nil
    var domain: String? = nil
    
    var resolved: Bool {
        get {
            return address != nil
        }
    }
    
    private var netService: NSNetService? {
        willSet {
            netService?.delegate = nil
        }
        
        didSet {
            netService?.delegate = self
        }
    }
    
    override init() {
        super.init()
    }
    
    convenience init(domain dom: String?, type typ: String?, name nam: String?) {
        self.init()
        
        domain = dom
        type = typ
        name = nam
    }
    
    convenience init(netService service: NSNetService) {
        self.init(domain: service.domain, type: service.type, name: service.name)
        
        /*
         * Cannot assign netService directly in constructor
         * otherwise didSet isn't called
         * http://stackoverflow.com/questions/25230780/is-it-possible-to-allow-didset-to-be-called-during-initialization-in-swift
         */
        setup(service)
    }
    
    private func setup(service: NSNetService) {
        netService = service
    }
    
    func resolve() {
        if netService == nil {
            netService = NSNetService(domain: domain!, type: type!, name: name!)
        }
        readResolvedInfo()
        if address == nil {
            netService?.resolveWithTimeout(Gateway.ResolveTimeout)
        }
    }
    
    func readResolvedInfo() {
        if let service = netService {
            port = service.port
            hostname = service.hostName
            if let addr = service.addresses {
                if addr.count > 0 {
                    let data = addr[0]
                    var storage = sockaddr_storage()
                    data.getBytes(&storage, length: sizeof(sockaddr_storage))
                    
                    let family = Int32(storage.ss_family)
                    switch(family) {
                    case AF_INET:
                        withUnsafePointer(&storage) {
                            address = String(CString: inet_ntoa(UnsafePointer<sockaddr_in>($0).memory.sin_addr), encoding: NSASCIIStringEncoding)
                        }
                        break;
                    case AF_INET6:
                        withUnsafePointer(&storage) {
                            var addr6 = UnsafePointer<sockaddr_in6>($0).memory.sin6_addr
                            withUnsafePointer(&addr6) {
                                var ip6AddressString = [CChar](count:Int(INET6_ADDRSTRLEN), repeatedValue: 0)
                                inet_ntop(
                                    AF_INET6 ,
                                    UnsafePointer<Void>($0),
                                    &ip6AddressString,
                                    socklen_t(INET6_ADDRSTRLEN))
                                address = String(CString: ip6AddressString, encoding: NSASCIIStringEncoding)
                            }
                        }
                        break;
                    default:
                        break;
                    }
                }
            }
        }
    }
    
    // MARK: - NSNetServiceDelegate
    
    /* Sent to the NSNetService instance's delegate prior to resolving a service on the network. If for some reason the resolution cannot occur, the delegate will not receive this message, and an error will be delivered to the delegate via the delegate's -netService:didNotResolve: method.
    */
    func netServiceWillResolve(sender: NSNetService) {
        DDLogVerbose("netServiceWillResolve");
    }
    
    /* Sent to the NSNetService instance's delegate when one or more addresses have been resolved for an NSNetService instance. Some NSNetService methods will return different results before and after a successful resolution. An NSNetService instance may get resolved more than once; truly robust clients may wish to resolve again after an error, or to resolve more than once.
    */
    func netServiceDidResolveAddress(sender: NSNetService) {        
        readResolvedInfo()
        netService = nil
        DDLogVerbose("netServiceDidResolveAddress: \(address):\(port)")
    }
    
    /* Sent to the NSNetService instance's delegate when an error in resolving the instance occurs. The error dictionary will contain two key/value pairs representing the error domain and code (see the NSNetServicesError enumeration above for error code constants).
    */
    func netService(sender: NSNetService, didNotResolve errorDict: [String : NSNumber]) {
        DDLogVerbose("netService didNotResolve: \(errorDict.description)");
        netService = nil
    }
    
    /* Sent to the NSNetService instance's delegate when the instance's previously running publication or resolution request has stopped.
    */
    func netServiceDidStop(sender: NSNetService) {
        DDLogVerbose("netServiceDidStop");
        netService = nil
    }
    
}