//
//  ConnectionKeeper.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import Foundation

@objc public protocol ConnectionKeeperDelegate {
    func doConnect()
    func doDisconnect()
    func isConnected() -> Bool
    func isConnecting() -> Bool
    func isDisconnecting() -> Bool
}

@objc public class ConnectionKeeper: NSObject {
    let target : String?
    let reachability : Reachability
    var backOff: ExponentialBackOff
    var delegate: ConnectionKeeperDelegate?
    var started = false
    private var timer: NSTimer? = nil
    private let dispatchQueue: dispatch_queue_t
    
    public init(delegate: ConnectionKeeperDelegate, target: String?) {
        self.target = target
        if let server = target {
            if !server.isEmpty {
                reachability = Reachability(hostname: server)
            } else {
                reachability = Reachability.reachabilityForInternetConnection()
            }
        } else {
            reachability = Reachability.reachabilityForInternetConnection()
        }
        
        self.delegate = delegate
        backOff = ExponentialBackOff()
        dispatchQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_BACKGROUND, 0)
        super.init()
        
        reachability.reachableBlock = { (reachability: Reachability?) -> () in
            dispatch_async(self.dispatchQueue, { () -> () in
                if let del = self.delegate {
                    if(del.isConnected() || del.isConnecting()) {
                        return
                    }
                self.cancelRetry()
                self.backOff.reset()
                self.connect()
                }
            })
        }
        reachability.unreachableBlock = { (reachability: Reachability?) -> () in
            dispatch_async(self.dispatchQueue, { () -> () in
                self.cancelRetry()
                self.disconnect()
            })
        }
    }
    
    public convenience init(delegate: ConnectionKeeperDelegate) {
        self.init(delegate: delegate, target: nil)
    }
    
    public func start() {
        self.started = true
        dispatch_async(dispatchQueue, { () -> () in
            DDLogVerbose("ConnectionKeeper: start")
            self.reachability.startNotifier()
        })
    }
    
    public func stop() {
        self.started = false
        dispatch_async(dispatchQueue, { () -> () in
            DDLogVerbose("ConnectionKeeper: stop")
            self.cancelRetry()
            self.reachability.stopNotifier()
        })
    }
    
    public func onConnectSuccess() {
        dispatch_async(dispatchQueue, { () -> () in
            DDLogVerbose("ConnectionKeeper: onConnectSuccess")
            self.cancelRetry()
            self.backOff.reset()
        })
    }
    
    public func onConnectFailure() {
        dispatch_async(dispatchQueue, { () -> () in
            DDLogVerbose("ConnectionKeeper: onConnectFailure")
            self.retryConnect()
        })
    }
    
    public func onDisconnectSuccess() {
        DDLogVerbose("ConnectionKeeper: onDisconnectSuccess")
    }
    
    public func onDisconnectFailure() {
        DDLogVerbose("ConnectionKeeper: onDisconnectFailure")
    }
    
    public func onConnectionLost() {
        dispatch_async(dispatchQueue, { () -> () in
            DDLogVerbose("ConnectionKeeper: onConnectionLost")
            self.backOff.reset()
            self.retryConnect()
        })
    }
    
    func connect() {
        DDLogVerbose("ConnectionKeeper: connect")
        delegate?.doConnect()
    }
    
    func disconnect() {
        DDLogVerbose("ConnectionKeeper: disconnect")
        delegate?.doDisconnect()
    }
    
    func retryConnect() {
        if !started {
            return
        }
        if !reachability.isReachable() {
            return
        }
        cancelRetry()
        let delay = backOff.nextBackOffInterval()
        DDLogVerbose("ConnectionKeeper: retry after \(delay)")
        let tm = NSTimer(timeInterval: delay, target: self, selector: "connect", userInfo: nil, repeats: false)
        timer = tm
        NSRunLoop.mainRunLoop().addTimer(tm, forMode: NSRunLoopCommonModes)
    }
    
    func cancelRetry() {
        timer?.invalidate()
        timer = nil
    }
}