//
//  Phone.swift
//  VMS
//
//  Copyright (c) 2015 Skymotion. All rights reserved.
//

import UIKit

class Phone: NSObject, AbtoPhoneInterfaceObserver, ConnectionKeeperDelegate {
    
    struct StatusCode {
        static let Forbidden: Int32 = 403
    }
    
    /*
     * Make enum observable using KVO:
     * http://stackoverflow.com/a/31947280
     */
    @objc
    enum RegistrationStatus: Int {
        case Unregistered
        case Registering
        case Registered
        case Unregistering
    }
    
    private static var _instance: Phone? = nil
    var phone: AbtoPhoneInterface = AbtoPhoneInterface()
    var keeper: ConnectionKeeper? = nil
    var gateway: Gateway? = nil
    dynamic var status: RegistrationStatus = .Unregistered
    var configured: Bool {
        get {
            var isConfigured = false
            if let cfg = phone.config() {
                if let usr = cfg.regUser {
                    isConfigured = !usr.isEmpty
                }
                if let pwd = cfg.regPassword {
                    isConfigured = isConfigured && !pwd.isEmpty
                }
                if let domain = cfg.regDomain {
                    isConfigured = isConfigured && !domain.isEmpty
                }
            }
            return isConfigured
        }
    }
    
    class var sharedInstance: Phone {
        if _instance == nil {
            _instance = Phone()
        }
        return _instance!
    }
    
    override init() {
        super.init()
        
        phone.initialize(self)
        setupConfig()
        
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "applicationDidEnterBackground:", name: UIApplicationDidEnterBackgroundNotification, object: nil)
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "applicationWillEnterForeground:", name: UIApplicationWillEnterForegroundNotification, object: nil)
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "applicationWillTerminate:", name: UIApplicationWillTerminateNotification, object: nil)
    }
    
    func start() {
        guard configured else {
            return
        }
        
        connect()
        if !keeper!.started {
            keeper?.start()
        }
    }
    
    func stop() {
        keeper?.stop()
        keeper = nil
        disconnect()
    }
    
    func restart() {
        self.stop()
        self.start()
    }
    
    func connect() {
        status = .Registering
        if keeper == nil {
            let config = phone.config()
            if let domain = config.regDomain {
                keeper = ConnectionKeeper(delegate: self, target: domain)
            } else {
                keeper = ConnectionKeeper(delegate: self)
            }
        }
        phone.finalizeConfiguration()
    }
    
    func disconnect() {
        let registered = status == .Registered
        status = .Unregistering
        if registered {
            phone.unregister()
        }
    }
    
    func answerCall(callId: Int) {
        phone.answerCall(callId, status: 200, withVideo: false)
    }
    
    func hangUpCall(callId: Int) {
        phone.hangUpCall(callId, status: 486)
    }
    
    func startCall(number: String) -> Int {
        if status == .Registered {
            let callId = phone.startCall(number, withVideo: false)
            CallLog.sharedInstance.startCall(callId, number: number, incoming: false)
            return callId
        } else {
            return kInvalidCallId
        }
    }
    
    func saveSettings() {
        savePhoneConfig()
        saveGateway()
    }
    
    func savePhoneConfig() {
        let config = phone.config()
        config.saveToUserDefaults(SETTINGS_KEY)
    }
    
    func saveGateway() {
        if let gt = gateway {
            GatewayManager.sharedInstance.saveGatewayToUserDefaults(gt)
        }
    }
    
    func loadGateway() {
        gateway = GatewayManager.sharedInstance.loadGatewayFromUserDefaults()
        gateway?.resolve()
    }
    
    func setupConfig() {
        let config = phone.config()
        config.setFromConfig(AbtoPhoneConfig.loadFromUserDefaults(SETTINGS_KEY))
        
        config.enableRingBackTone = true
        config.enableRingTone = true
        let ring = "abto_ringtone.mp3"
        config.ringBackToneUrl = String(format: "%@/%@", NSBundle.mainBundle().resourcePath!, ring)
        config.ringToneUrl = config.ringBackToneUrl
        
        config.regAuthId = nil
        config.displayName = nil
        
        config.enableSrtp = false
        config.stun = "stun.l.google.com:19302"
        /*
        stun:stun.l.google.com:19302
        stun:stun1.l.google.com:19302
        stun:stun2.l.google.com:19302
        stun:stun3.l.google.com:19302
        stun:stun4.l.google.com:19302
        */
        
        config.signalingTransport = PhoneSignalingTransport.Udp
        savePhoneConfig()
        
        loadGateway()
    }
    
    func setCredentials(user: String?, password: String?) {
        setCredentials(user, password: password, server: nil)
    }
    
    func setCredentials(user: String?, password: String?, server: String?) {
        let config = phone.config()
        
        config.regUser = user
        config.regPassword = password
        config.regDomain = server
        config.proxy = server
    }
    
    func getCredentials() -> (String?, String?) {
        let config = phone.config()
        
        return (config.regUser, config.regPassword)
    }
    
    func getServer() -> String? {
        let config = phone.config()
        
        return config.regDomain
    }

    func setServer(server: String?) {
        let config = phone.config()
        
        config.regDomain = server
        config.proxy = server
    }
    
    func applicationDidEnterBackground(notification: NSNotification) {
        // Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later.
        // If your application supports background execution, this method is called instead of applicationWillTerminate: when the user quits.
        phone.keepAwake(true)
    }
    
    func applicationWillEnterForeground(notification: NSNotification) {
        // Called as part of the transition from the background to the inactive state; here you can undo many of the changes made on entering the background.
        
        phone.keepAwake(false)
    }
    
    func applicationWillTerminate(notification: NSNotification) {
        // Called when the application is about to terminate. Save data if appropriate. See also applicationDidEnterBackground:.
        NSNotificationCenter.defaultCenter().removeObserver(self)
        phone.deinitialize()
    }
    
    
    // MARK: ConnectionKeeperDelegate
    
    func doConnect() {
        dispatch_async(dispatch_get_main_queue(), {
            self.connect()
        })
    }
    
    func doDisconnect() {
        dispatch_async(dispatch_get_main_queue(), {
            self.disconnect()
        })
    }
    
    func isConnected() -> Bool {
        return status == .Registered
    }
    
    func isConnecting() -> Bool {
        return status == .Registering
    }
    
    func isDisconnecting() -> Bool {
        return status == .Unregistering
    }
    
    // MARK: - Abto SIP SDK delegate
    
    func onRegistered(accId: Int) {
        DDLogVerbose("Phone onRegistered: \(accId)")
        status = .Registered
        keeper?.onConnectSuccess()
        saveSettings()
        
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_PHONE_EVENT, object: NSNumber(integer: PhoneEvents.RegSuccess.rawValue))
    }
    
    func onRegistrationFailed(accId: Int, statusCode: Int32, statusText: String!) {
        DDLogVerbose("Phone onRegistrationFailed: \(accId), status \(statusCode), \(statusText)")
        if statusCode == StatusCode.Forbidden {
            keeper?.stop()
            dispatch_async(dispatch_get_main_queue(), {
                let alert = UIAlertView(title: NSLocalizedString("phone.registration.forbidden.title", comment: ""),
                    message: NSLocalizedString("phone.registration.forbidden.message", comment: ""),
                    delegate: nil,
                    cancelButtonTitle: NSLocalizedString("ok", comment: ""))
                alert.show()
            })
        } else {
            if status == .Registered {
                keeper?.onConnectionLost()
            } else {
                keeper?.onConnectFailure()
            }
        }
        status = .Unregistered
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_PHONE_EVENT, object: NSNumber(integer: PhoneEvents.RegFailed.rawValue))
    }
    
    func onUnRegistered(accId: Int) {
        DDLogVerbose("Phone onUnRegistered: \(accId)")
        keeper?.onDisconnectSuccess()
        status = .Unregistered
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_PHONE_EVENT, object: NSNumber(integer: PhoneEvents.UnregSuccess.rawValue))
    }
    
    func onRemoteAlerting(accId: Int, statusCode: Int32) {
        DDLogVerbose("Phone onRemoteAlerting: \(accId), status \(statusCode)")
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_PHONE_EVENT, object: NSNumber(integer: PhoneEvents.RemoteAlerting.rawValue))
    }
    
    func onIncomingCall(callId: Int, remoteContact: String!) {
        DDLogVerbose("Phone onIncomingCall: \(callId), remoteContact \(remoteContact)")
        CallLog.sharedInstance.startCall(callId, number: remoteContact, incoming: true)
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_CALL_EVENT, object: NSNumber(integer: CallEvents.Incoming.rawValue), userInfo: [
            CALL_ID_ARGUMENT: NSNumber(integer: callId),
            CONTACT_ARGUMENT: remoteContact
            
            ])
    }
    
    func onCallConnected(callId: Int, remoteContact: String!) {
        DDLogVerbose("Phone onCallConnected: \(callId), remoteContact \(remoteContact)")
        CallLog.sharedInstance.connectCall(callId)
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_CALL_EVENT, object: NSNumber(integer: CallEvents.Connected.rawValue), userInfo: [
            CALL_ID_ARGUMENT: NSNumber(integer: callId)
            ])
    }
    
    func onCallDisconnected(callId: Int, remoteContact: String!, statusCode: Int, message: String!) {
        DDLogVerbose("Phone onCallDisconnected: \(callId), remoteContact \(remoteContact), status \(statusCode), \(message)")
        CallLog.sharedInstance.disconnectCall(callId)
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_CALL_EVENT, object: NSNumber(integer: CallEvents.Disconnected.rawValue), userInfo: [
            CALL_ID_ARGUMENT: NSNumber(integer: callId),
            STATUS_ARGUMENT: NSNumber(integer: statusCode)
            ])
    }
    
    func onCallAlerting(callId: Int, statusCode: Int32) {
        DDLogVerbose("Phone onCallAlerting: \(callId), status \(statusCode)")
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_CALL_EVENT, object: NSNumber(integer: CallEvents.Alerting.rawValue), userInfo: [
            CALL_ID_ARGUMENT: NSNumber(integer: callId),
            STATUS_ARGUMENT: NSNumber(int: statusCode)
            ])
    }
    
    func onPresenceChanged(uri: String!, status: PhoneBuddyStatus, note: String!) {
        DDLogVerbose("Phone onPresenceChanged: \(uri), status \(status), \(note)")
    }
    
    func onCallHeld(callId: Int, state: Bool) {
        DDLogVerbose("Phone onCallHeld: \(callId), state \(state)")
    }
    
    func onToneReceived(callId: Int, tone: Int) {
        DDLogVerbose("Phone onToneReceived: \(callId), tone \(tone)")
    }
    
    func onTextMessageReceived(from: String!, to: String!, body: String!) {
        DDLogVerbose("Phone onTextMessageReceived: from \(from), to \(to), body \(body)")
    }
    
    func onTextMessageStatus(address: String!, reason: String!, status: Bool) {
        DDLogVerbose("Phone onTextMessageStatus: address \(address), reason \(reason), status \(status)")
    }
    
    func onTransferStatus(callId: Int, statusCode: Int32, message: String!) {
        DDLogVerbose("Phone onTransferStatus: callId \(callId), status \(statusCode), \(message)")
        NSNotificationCenter.defaultCenter().postNotificationName(NOTIFICATION_CALL_EVENT, object: NSNumber(integer: CallEvents.Transfering.rawValue), userInfo: [
            CALL_ID_ARGUMENT: NSNumber(integer: callId),
            STATUS_ARGUMENT: NSNumber(int: statusCode),
            MESSAGE_ARGUMENT: message
            ])
    }
}